
import numpy as np 
import pandas as pd
import matplotlib.pyplot as plt 



plt.rcParams['figure.figsize'] = (9,6)
plt.rcParams['font.size'] = 14
plt.rcParams['legend.fontsize'] = 24
#plt.rcParams['text.usetex'] = True
#plt.rcParams['font.family'] = 'serif'


SCHEDULE = 25
ARMS = 100

CASE = 1
BATCHSIZE = 5
REINFORCELR = 0.001
numClass1 = 2
numClass2 = 2
EPISODEEND = 1000000
INTERVAL = 10000
RUNS = 200

NOISY = True

print(f'Selected case: {CASE}')
numEpisode = np.arange(0, EPISODEEND + INTERVAL, INTERVAL) 


reinforceRewards = []
nnRewards = []



sizeAwareFileName = (f'../testResults/size_aware_env/case_{CASE}/sizeAwareIndexResults_arms_{ARMS}_schedule_{SCHEDULE}_arms.csv')
df = pd.read_csv(sizeAwareFileName)
sizeAwareRewards = df.iloc[:,1]

sizeAware5Percentile, sizeAware95Percentile = np.percentile(sizeAwareRewards, [5, 95])
sizeAwareRewards = np.ceil(sum(sizeAwareRewards[0:RUNS]) / RUNS)


for i in range(RUNS):
    if NOISY:
        NeurWINFileName = (f'../testResults/size_aware_env/noisy_results/case_{CASE}/nnIndexResults_arms_{ARMS}_batchSize_{BATCHSIZE}_run_{i}_schedule_{SCHEDULE}_arms.csv')
    else:
        NeurWINFileName = (f'../testResults/size_aware_env/case_{CASE}/nnIndexResults_arms_{ARMS}_batchSize_{BATCHSIZE}_run_{i}_schedule_{SCHEDULE}_arms.csv')
    df = pd.read_csv(NeurWINFileName)
    runReward = df.iloc[:, 1]
    nnRewards.append(runReward)

nnRewards = np.array(nnRewards)
nnVal = np.ceil(np.sum(nnRewards, 0) / RUNS)


nnRewards = np.transpose(nnRewards)


percentile5 = np.percentile(nnRewards, 5, axis=1)
percentile95 = np.percentile(nnRewards, 95, axis=1)

'''
for i in range(RUNS):
    reinforceFileName = (f'../testResults/size_aware_env/case_{CASE}/reinforceResults_arms_{ARMS}\
_batchSize_{BATCHSIZE}_lr_{REINFORCELR}_run_{i}_schedule_{SCHEDULE}.csv')
    df = pd.read_csv(reinforceFileName)
    runCost = df.iloc[:, 1]
    reinforceRewards.append(runCost)


reinforcePercentileRewards = np.transpose(reinforceRewards)
reinforceRewards = np.ceil(sum(reinforceRewards))  / RUNS

percentile5Reinforce = np.percentile(reinforcePercentileRewards, 5, axis=1)
percentile95Reinforce = np.percentile(reinforcePercentileRewards, 95, axis=1)
'''


plotNumEpisodesXAxis = np.arange(0, EPISODEEND + 100000, 100000) 

plt.plot(numEpisode, nnVal, label=f'NeurWIN', color='b', linewidth=3.0, linestyle='solid')
#plt.plot(numEpisode, reinforceRewards, label='REINFORCE', color='k', linewidth=3.0, linestyle='dotted')
plt.hlines(xmin=0, xmax=EPISODEEND, y=sizeAwareRewards, label='Size-aware Index', color='r', linewidth=3.0, linestyle='dashdot')

plt.fill_between(x=numEpisode,y1=sizeAware5Percentile, y2=sizeAware95Percentile, alpha=0.3, color='orange')#, label='Size-aware Confidence Bound')
plt.fill_between(x=numEpisode,y1=percentile5, y2=percentile95, alpha=0.3, color='green')#, label='NeurWIN Confidence Bound')
#plt.fill_between(x=numEpisode,y1=percentile5Reinforce, y2=percentile95Reinforce, alpha=0.3, color='slateblue')#, label='REINFORCE Confidence Bound')



#plt.xticks(plotNumEpisodesXAxis)

plt.legend()
plt.xlabel('Number of Training Episodes', fontsize=24)
plt.ylabel(f'Total Discounted Rewards', fontsize=24)
plt.yticks(rotation=60)

if NOISY:
    plt.savefig(f'../plotResults/size_aware_results/noisy_results/case{CASE}Results_arms_{ARMS}_mini_batch_{BATCHSIZE}_schedule_{SCHEDULE}_arms.pdf')
else:
    plt.savefig(f'../plotResults/size_aware_results/case{CASE}Results_arms_{ARMS}_mini_batch_{BATCHSIZE}_schedule_{SCHEDULE}_arms.pdf')

plt.show()



'''
############# Plotting relative reward #############


relativeReward = nnVal - sizeAwareRewards 

percentileRelative5 = np.percentile(relativeReward, 5, axis=0)
percentileRelative95 = np.percentile(relativeReward, 95, axis=0)


plt.plot(numEpisode, relativeReward, label=f'Reward Difference', color='b', linewidth=3.0, linestyle='solid')
#plt.fill_between(x=numEpisode,y1=percentileRelative5, y2=percentileRelative5, alpha=0.3, label='Confidence Bound')

#plt.xticks(plotNumEpisodesXAxis)

plt.legend()
plt.xlabel('Number of Training Episodes', fontsize=24)
plt.ylabel(f'Relative Average Reward To Baseline', fontsize=24)
plt.yticks(rotation=60)

#plt.title('CASE: {}. Cumulative Discounted Holding Cost for {} jobs: {} class 1. {} class 2. Schedule {} Arms at Each Timeslot'.format(CASE, ARMS, numClass1, numClass2, SCHEDULE))
plt.savefig(f'../plotResults/size_aware_results/relative_rewards_case{CASE}Results_arms_{ARMS}_mini_batch_{BATCHSIZE}_schedule_{SCHEDULE}_arms.pdf')

plt.show()


'''





